
import logging
import unittest

import mock

import swailing.logger
import swailing.token_bucket
from swailing.token_bucket import TokenBucket


class TokenBucketTest(unittest.TestCase):
    """TokenBucket test.

    Since TokenBucket depends on time.time(), we mock it out. The
    `clock.return_value = X` statements set the clock to that time.

    """

    @mock.patch('swailing.token_bucket.time.time')
    def test_simple(self, clock):
        clock.return_value = 1
        tb = TokenBucket(10, 100)
        self.assertEqual(len(tb), 100)
        self.assertTrue(tb.check_and_consume())

    @mock.patch('swailing.token_bucket.time.time')
    def test_throttle(self, clock):
        clock.return_value = 1
        tb = TokenBucket(10, 100)

        # we get to consume 100 tokens...
        for i in xrange(100):
            self.assertTrue(tb.check_and_consume())

        # ...then next one fails
        self.assertFalse(tb.check_and_consume())

        # fill halfway and check level
        clock.return_value += 0.5
        tb._fill()
        self.assertEqual(len(tb), 5)

        clock.return_value += 0.5
        tb._fill()
        self.assertEqual(len(tb), 10)

        # this fill maxes out capacity
        clock.return_value += 10
        tb._fill()
        self.assertEqual(len(tb), 100)

        # attempt to consume more than allowed then check throttle
        # count
        for i in xrange(123):
            tb.check_and_consume()
        self.assertEqual(tb.throttle_count, 23)

        # fill, consume, then check throttle count again
        clock.return_value += 1
        tb.check_and_consume()
        self.assertEqual(tb.throttle_count, 0)


class LoggerTest(unittest.TestCase):
    def test_fallback(self):
        mock_logger = mock.Mock()
        logger = swailing.Logger(mock_logger, 10, 100)

        logger.info("hello world!")
        mock_logger.log.assert_called_with(
            logging.INFO,
            "hello world!",
        )

    def test_context(self):
        mock_logger = mock.Mock()
        logger = swailing.Logger(mock_logger, 10, 100)
        with logger.info() as L:
            L.primary("primary")
            L.detail("detail")
            L.hint("hint")

        calls = [
            mock.call(logging.INFO, "primary"),
            mock.call(logging.INFO, "detail"),
            mock.call(logging.INFO, "hint"),
        ]
        mock_logger.log.assert_has_calls(calls)

    def test_no_rate(self):
        """Simple test without rate limiting."""

        mock_logger = mock.Mock()
        logger = swailing.Logger(mock_logger)
        with logger.info() as L:
            L.primary("primary")
            L.detail("detail")
            L.hint("hint")

        calls = [
            mock.call(logging.INFO, "primary"),
            mock.call(logging.INFO, "detail"),
            mock.call(logging.INFO, "hint"),
        ]
        mock_logger.log.assert_has_calls(calls)

    @mock.patch('swailing.token_bucket.time.time')
    def test_throttle(self, clock):
        mock_logger = mock.Mock()
        clock.return_value = 1
        logger = swailing.Logger(mock_logger, 1, 2)

        # We'll write a buch of times, but since capacity is just 2,
        # we should expect the root_logger to only have been called
        # twice.
        for i in xrange(10):
            with logger.info() as L:
                L.primary("primary %d" % i)
                L.detail("detail %d" % i)
                L.hint("hint %d" % i)

        calls = [
            mock.call(logging.INFO, "primary 0"),
            mock.call(logging.INFO, "detail 0"),
            mock.call(logging.INFO, "hint 0"),

            mock.call(logging.INFO, "primary 1"),
            mock.call(logging.INFO, "detail 1"),
            mock.call(logging.INFO, "hint 1"),
        ]
        mock_logger.log.assert_has_calls(calls)


        # Now let's wait a second so we get one more token, then run test again.
        clock.return_value += 1
        mock_logger.reset_mock()

        for i in xrange(10):
            with logger.info() as L:
                L.primary("primary %d" % i)
                L.detail("detail %d" % i)
                L.hint("hint %d" % i)

        calls = [
            mock.call(logging.INFO, "primary 0"),
            mock.call(logging.INFO, "detail 0"),
            mock.call(logging.INFO, "hint 0"),
        ]
        mock_logger.log.assert_has_calls(calls)


        # Force a throttle message.
        clock.return_value += 100
        mock_logger = mock.Mock()
        logging.basicConfig()
        logger = swailing.Logger(mock_logger, 1, 1)
        for i in xrange(10):
            logger.error("foo")
        clock.return_value += 100
        logger.error("bar")

        calls = [
            mock.call(logging.ERROR, "foo"),
            mock.call(logging.ERROR, ""),
            mock.call(logging.ERROR, "(... throttled %d messages ...)", 9),
            mock.call(logging.ERROR, ""),
            mock.call(logging.ERROR, "bar"),
        ]
        mock_logger.log.assert_has_calls(calls)

    def test_verbosity(self):
        mock_logger = mock.Mock()
        logger = swailing.Logger(mock_logger, 10, 100, verbosity=swailing.PRIMARY)
        with logger.info() as L:
            L.primary("primary")
            L.detail("detail")
            L.hint("hint")

        calls = [
            mock.call(logging.INFO, "primary"),
        ]
        self.assertEqual(mock_logger.log.mock_calls, calls)

        mock_logger = mock.Mock()
        logger = swailing.Logger(mock_logger, 10, 100, verbosity=swailing.DETAIL)
        with logger.info() as L:
            L.primary("primary")
            L.detail("detail")
            L.hint("hint")

        calls = [
            mock.call(logging.INFO, "primary"),
            mock.call(logging.INFO, "detail"),
        ]
        self.assertEqual(mock_logger.log.mock_calls, calls)

        mock_logger = mock.Mock()
        logger = swailing.Logger(mock_logger, 10, 100, verbosity=swailing.HINT)
        with logger.info() as L:
            L.primary("primary")
            L.detail("detail")
            L.hint("hint")

        calls = [
            mock.call(logging.INFO, "primary"),
            mock.call(logging.INFO, "detail"),
            mock.call(logging.INFO, "hint"),
        ]
        self.assertEqual(mock_logger.log.mock_calls, calls)




import logging
import threading
from contextlib import contextmanager

from swailing.token_bucket import TokenBucket


PRIMARY = 10
DETAIL = 20
HINT = 30


class Logger(object):
    """A logging.Logger-like object with some swailing goodness.

    """

    def __init__(self, name_or_logger, fill_rate=None, capacity=None, verbosity=HINT):
        """Set up Logger-like object with some swailing goodness.

        name_or_logger is either a (possibly unicode) string or a
        logging.Logger-like object. If supplied as a string, we simply
        use logging.getLogger.

        fill_rate is the number of tokens per second Logger
        accumulates. Each log output consumes one token.

        capacity is the maximum number of tokens the Logger is allowed
        to have accumulated.

        verbosity is the initial verbosity level (defaults to HINT).
        Setting it to PRIMARY will stop detail and hint outputs.
        Setting it to DETAIL will stop hint outputs. Setting it to
        HINT allows all output.

        """

        if fill_rate and capacity:
            self._tb = TokenBucket(fill_rate, capacity)
        else:
            self._tb = None
        self._tb_lock = threading.Lock()

        if isinstance(name_or_logger, basestring):
            self._logger = logging.getLogger(name_or_logger)
        else:
            self._logger = name_or_logger

        self._verbosity = verbosity

    def debug(self, msg=None, *args, **kwargs):
        """Write log at DEBUG level. Same arguments as Python's built-in
        Logger.

        """

        return self._log(logging.DEBUG, msg, args, kwargs)

    def info(self, msg=None, *args, **kwargs):
        """Similar to DEBUG but at INFO level."""

        return self._log(logging.INFO, msg, args, kwargs)

    def warning(self, msg=None, *args, **kwargs):
        """Similar to DEBUG but at WARNING level."""

        return self._log(logging.WARNING, msg, args, kwargs)

    def error(self, msg=None, *args, **kwargs):
        """Similar to DEBUG but at ERROR level."""

        return self._log(logging.ERROR, msg, args, kwargs)

    def critical(self, msg=None, *args, **kwargs):
        """Similar to DEBUG but at CRITICAL level."""

        return self._log(logging.CRITICAL, msg, args, kwargs)

    def log(self, level, msg=None, *args, **kwargs):
        """Writes log out at any arbitray level."""

        return self._log(level, msg, args, kwargs)

    def set_verbosity(self, level):
        self._verbosity = level


    def _log(self, level, msg, args, kwargs):
        """Throttled log output."""

        with self._tb_lock:
            if self._tb is None:
                throttled = 0
                should_log = True
            else:
                throttled = self._tb.throttle_count
                should_log = self._tb.check_and_consume()

        if should_log:
            if throttled > 0:
                self._logger.log(level, "")
                self._logger.log(
                    level,
                    "(... throttled %d messages ...)",
                    throttled,
                )
                self._logger.log(level, "")

            if msg is not None:
                self._logger.log(level, msg, *args, **kwargs)

            return FancyLogContext(self._logger, level, self._verbosity)
        else:
            return NoopLogContext()


class FancyLogContext(object):
    """Accepts primary, detail, and hint log statements and outputs them
    when the context exits without error.

    """

    def __init__(self, logger, level, verbosity):
        self._logger = logger
        self._level = level
        self._verbosity = verbosity
        self._log = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # do not log anything if an exception occurs while forming the
        # log messages
        if exc_type or exc_value or traceback:
            return

        for msg, args, kwargs in self._log.values():
            self._logger.log(self._level, msg, *args, **kwargs)

    def primary(self, msg, *args, **kwargs):
        if self._verbosity >= PRIMARY:
            self._log[PRIMARY] = (msg, args, kwargs)

    def detail(self, msg, *args, **kwargs):
        if self._verbosity >= DETAIL:
            self._log[DETAIL] = (msg, args, kwargs)

    def hint(self, msg, *args, **kwargs):
        if self._verbosity >= HINT:
            self._log[HINT] = (msg, args, kwargs)


class NoopLogContext(object):
    """Fakes a FancyLogContext but does nothing."""

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def primary(self, msg, *args, **kwargs):
        pass

    def detail(self, msg, *args, **kwargs):
        pass

    def hint(self, msg, *args, **kwargs):
        pass



from setuptools import setup

setup(
    name="swailing",
    version="0.1.1",
    packages=['swailing'],
    author="Wes Chow",
    author_email="wes@chartbeat.com",
    description=("Logging library for apps that produce debilitating amounts of log output."),
    license="Apache",
    keywords="logging",
    url="https://github.com/chartbeat-labs/swailing",
    long_description="""
An opinionated logging library for applications that produce a debilitating amount of log output.

Swailing uses the token bucket algorithm and provides facilities for
PostgreSQL style error logs. See
https://github.com/chartbeat-labs/swailing for more details.
""",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Utilities",
        "Topic :: System :: Logging",
        "Programming Language :: Python",
        "License :: OSI Approved :: Apache Software License",
    ],
)




import time


class TokenBucket(object):
    """Implements token bucket algorithm.

    https://en.wikipedia.org/wiki/Token_bucket
    """

    def __init__(self, fill_rate, capacity):
        self._fill_rate = float(fill_rate)
        self._capacity = float(capacity)
        self._count = float(capacity)
        self._last_fill = time.time()

        self.throttle_count = 0

    def check_and_consume(self):
        """Returns True if there is currently at least one token, and reduces
        it by one.

        """

        if self._count < 1.0:
            self._fill()

        consumable = self._count >= 1.0
        if consumable:
            self._count -= 1.0
            self.throttle_count = 0
        else:
            self.throttle_count += 1

        return consumable

    def __len__(self):
        """Returns current number of discrete tokens."""

        return int(self._count)

    def _fill(self):
        """Fills bucket with accrued tokens since last fill."""

        right_now = time.time()
        time_diff = right_now - self._last_fill
        if time_diff < 0:
            return

        self._count = min(
            self._count + self._fill_rate * time_diff,
            self._capacity,
        )
        self._last_fill = right_now
